import json
import os
import typing
import uuid
from dataclasses import dataclass
from functools import wraps
from pathlib import Path

DEFAULT_ENDPOINT = 'https://migas.herokuapp.com/graphql'
DEFAULT_CONFIG_FILE = Path.home() / '.cache' / 'migas' / 'config.json'

# TODO: 3.10 - Replace with | operator
File = typing.Union[str, Path]


def suppress_errors(func):
    """Decorator to silently fail the wrapped function"""

    @wraps(func)
    def safe(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            pass

    return safe


def telemetry_enabled(func: typing.Callable) -> typing.Callable:
    """Decorator function to verify telemetry collection is enabled."""

    @wraps(func)
    def can_send(*args, **kwargs):
        if not os.getenv("ENABLE_MIGAS", "0").lower() in ("1", "true", "y", "yes"):
            # do not communicate with server
            return {
                "success": False,
                "errors": [
                    {"message": "migas is not enabled - set ENABLE_MIGAS environment variable."}
                ],
            }
        # otherwise, ensure config is set up
        setup()
        return func(*args, **kwargs)

    return can_send


@dataclass(init=False, repr=False, eq=False)
class Config:
    """
    Class to store client-side configuration, facilitating communication with the server.

    The class stores the following components:
    - `endpoint`:
    URL of the graphql endpoint of the migas server.
    - `user_id`:
    A string representation of a UUID (RFC 4122) assigned to the user.
    - `session_id`:
    A string representation of a UUID assigned to the lifespan of the migas invocation.

    This class is not meant to be initialized, instead usage depends on class attributes.
    """

    endpoint: str = None
    user_id: str = None
    session_id: str = None
    _is_setup = False

    @classmethod
    def init(
        cls,
        *,
        endpoint: str = None,
        user_id: str = None,
        session_id: str = None,
        final: bool = True,
    ) -> None:
        """
        Setup migas configuration.

        If class was already configured, existing configuration is used.
        """
        if cls._is_setup:
            return
        if endpoint is not None:
            cls.endpoint = endpoint
        elif cls.endpoint is None:
            cls.endpoint = DEFAULT_ENDPOINT
        if user_id is not None or cls.user_id is None:
            try:
                uuid.UUID(user_id)
                cls.user_id = user_id
            except Exception:
                cls.user_id = gen_uuid()
        # Do not set automatically, leave to developers
        if session_id is not None:
            try:
                uuid.UUID(session_id)
                cls.session_id = session_id
            except Exception:
                pass
        cls._is_setup = final

    @classmethod
    @suppress_errors
    def load(cls, filename: File) -> bool:
        """Load existing configuration file, or create a new one."""
        config = json.loads(Path(filename).read_text())
        cls.init(final=False, **config)
        return True

    @classmethod
    @suppress_errors
    def save(cls, filename: File) -> str:
        """Save to a JSON file."""
        config = {field: getattr(cls, field) for field in cls.__annotations__.keys()}
        # TODO: Make safe when multiprocessing
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        Path(filename).write_text(json.dumps(config))
        return str(filename)

    @classmethod
    def _reset(cls):
        """Reset the config class attributes."""
        cls.endpoint = None
        cls.user_id = None
        cls.session_id = None
        cls._is_setup = False


def setup(
    *,
    endpoint: str = None,
    user_id: str = None,
    session_id: str = None,
    save_config: bool = True,
    filename: File = None,
) -> None:
    """
    Configure the client, and save configuration to an output file.

    This method is invoked before each API call, but can also be called by
    application developers for finer-grain control.
    """
    if Config._is_setup:
        return
    filename = filename or DEFAULT_CONFIG_FILE
    if Path(filename).exists():
        Config.load(filename)
    # if any parameters have been set, override the current attribute
    Config.init(endpoint=endpoint, user_id=user_id, session_id=session_id)
    if save_config:
        Config.save(filename)


def gen_uuid(uuid_factory: str = "safe") -> str:
    """
    Generate a RFC 4122 UUID.

    Depending on what `uuid_factory` is provided, the UUID will be generated differently:
    - `safe`: This is multiprocessing safe, and uses system information.
    - `random`: This is random, and may run into problems if setup is called across multiple
    processes.

    Hard cases to think about:
    - HPCs where HOSTNAME envvar is not set
    - Docker images where previous config is unavailable
    """
    # TODO: 3.10 - Replace with match/case
    if uuid_factory == "safe":
        return _safe_uuid_factory()
    elif uuid_factory == "random":
        return str(uuid.uuid4())
    raise NotImplementedError


def _safe_uuid_factory() -> str:
    import getpass
    import socket

    name = f"{getpass.getuser()}@{os.getenv('HOSTNAME', socket.gethostname())}"
    return str(uuid.uuid3(uuid.NAMESPACE_DNS, name))



"""Utility functions"""
from pathlib import Path
import platform
import sys


def is_ci():
    ...

def is_container():
    root = Path('/')
    if (root / '.dockerenv').exists():
        return 'docker'
    if (root / 'singularity').exists():
        return 'apptainer'
    return 'unknown'


def get_platform_info() -> dict:
    return {
        'python_version': platform.python_version(),
        'python_implementation': platform.python_implementation(),
        'platform': sys.platform,
    }


def compile_info() -> dict:
    data = get_platform_info()
    data['container'] = is_container()
    return data



"""
Create queries and mutations to be sent to the graphql endpoint.
"""
import sys
import typing
from uuid import UUID

from migas.config import Config, telemetry_enabled
from migas.request import request

if sys.version_info[:2] < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class OperationTemplate(TypedDict):
    operation: str
    args: dict


fromDateRange: OperationTemplate = {
    "operation": "query{from_date_range($)}",
    "args": {
        # required
        "project": '"{}"',
        "start": '"{}"',
        # optional
        "end": '"{}"',
        "unique": "{}",
    },
}


@telemetry_enabled
def from_date_range(
    project: str,
    start: str,
    end: str = None,
    unique: bool = False,
) -> dict:
    params = _introspec(from_date_range, locals())
    query = _formulate_query(params, fromDateRange)
    status, response = request(Config.endpoint, query)
    # TODO: Verify return is as expected
    return response


addProject: OperationTemplate = {
    "operation": "mutation{add_project(p:{$})}",
    "args": {
        # required
        "project": '"{}"',
        "project_version": '"{}"',
        "language": '"{}"',
        "language_version": '"{}"',
        # optional
        "status": "{}",
        "user": '"{}"',
        "session": '"{}"',
        "container": "{}",
        "platform": '"{}"',
        "arguments": '"{}"',
    },
}


@telemetry_enabled
def add_project(
    project: str,
    project_version: str,
    language: str,
    language_version: str,
    userId: UUID = None,
    sessionId: UUID = None,
    container: str = None,
    platform: str = None,
    arguments: list = None,
) -> dict:
    params = _introspec(add_project, locals())
    query = _formulate_query(params, addProject)
    status, response = request(Config.endpoint, query)
    return status, response


def _introspec(func: typing.Callable, func_locals: dict) -> dict:
    """Inspect a function and return all set parameters."""
    import inspect

    sig = inspect.signature(func)
    return {
        param: func_locals[param]
        for param, val in sig.parameters.items()
        if func_locals[param] != val.default
    }


def _formulate_query(params: dict, template: OperationTemplate) -> str:
    """Construct the graphql query."""
    qparams = {x: template["args"][x].format(params[x]) for x in template["args"] if x in params}
    query = template["operation"].replace("$", ",".join([f"{x}:{y}" for x, y in qparams.items()]))
    return query



from . import _version
__version__ = _version.get_versions()['version']

from .config import setup
from .operations import add_project, from_date_range

__all__ = (
    "__version__",
    "add_project",
    "from_date_range",
    "setup",
)



import uuid

import pytest

from .. import config


@pytest.fixture(autouse=True)
def tmp_config(tmp_path, monkeypatch):
    home = tmp_path / 'config.json'
    monkeypatch.setattr(config, 'DEFAULT_CONFIG_FILE', home)


def test_setup_default():

    conf = config.Config
    assert conf.endpoint is None
    assert conf.user_id is None
    assert conf.session_id is None
    assert conf._is_setup is False

    config.setup()
    assert conf.endpoint == config.DEFAULT_ENDPOINT
    assert uuid.UUID(conf.user_id)
    assert conf.session_id is None
    assert conf._is_setup is True

    # after being set up, cannot be overriden
    new_endpoint = 'https://github.com'
    config.setup(endpoint=new_endpoint)
    assert conf.endpoint == config.DEFAULT_ENDPOINT

    # but fine if cleared
    conf._reset()
    assert conf.endpoint is None
    config.setup(endpoint=new_endpoint)
    assert conf.endpoint == new_endpoint



"""Stripped down, minimal import way to communicate with server"""

import json
from http.client import HTTPConnection, HTTPSConnection, HTTPResponse
import typing
from urllib.parse import urlparse

from . import __version__


ETResponse = typing.Tuple[int, typing.Union[dict, str]]  # status code, body


def request(url: str, query: str, *, timeout: int = 3, chunk_size: int = None) -> ETResponse:
    purl = urlparse(url)
    # TODO: 3.10 - Replace with match/case
    if purl.scheme == 'https':
        Connection = HTTPSConnection
    elif purl.scheme == 'http':
        Connection = HTTPConnection
    else:
        raise ValueError("URL scheme not supported")

    body = json.dumps({"query": query}).encode("utf-8")
    conn = Connection(purl.netloc, timeout=timeout)
    headers = {
        'User-Agent': f'migas-client/{__version__}',
        'Accept-Encoding': 'gzip, deflate, br',
        'Accept': '*/*',
        'Content-Length': len(body),
        'Content-Type': 'application/json',
    }

    try:
        conn.request("POST", purl.path, body, headers)
        response = conn.getresponse()
        body = _read_response(response, chunk_size)
    except TimeoutError:
        return (
            408, {"success": False, "errors": [{"message": "Connection to server timed out."}]}
        )
    except ConnectionError:
        return (
            503, {"success": False, "errors": [{"message": "Server is not available."}]}
        )
    finally:
        conn.close()

    if not response.headers.get("X-Backend-Server"):
        # TODO: Implement logging
        print("migas server is incorrectly configured.")

    if response.headers.get("Content-Type").startswith("application/json"):
        body = json.loads(body)

    return response.status, body


def _read_response(response: HTTPResponse, chunk_size: int = None) -> str:
    """Read and aggregate the response body"""
    stream = b''
    # TODO: 3.8 - Replace with walrus
    # while chunk := response.read(chunk_size):
    chunk = 1
    while chunk:
        chunk = response.read(chunk_size)
        stream += chunk
    return stream.decode()
